//
// Created by philipp on 30.11.17.
//

#include <vector>
#include <algorithm>
#include <iostream>
#include "../include/PlatoonProtocolLib/Vehicle.h"
#include "../include/PlatoonProtocolLib/Utils.h"

using namespace platoonProtocol;

constexpr std::uint16_t Vehicle::UDP_PORT;
constexpr std::uint16_t Vehicle::TCP_PORT;
const PlatoonConfig Vehicle::DEFAULT_PLATOON_CONFIG{0.0f, 0.0f};

Vehicle::Vehicle(networking::Networking & net,
                 VehicleId ownVehicleId,
                 Vehicle::Role role)
    : net(net)
      , role(role)
      , ownEndpoint(createVehicleEndpoint(ownVehicleId, ""))
{
    platoonCreateServiceServer = networking::service::Server<PlatoonService>::create(
        net, ownEndpoint.getTcpPort(), PlatoonMessage::MAX_SIZE);
    platoonCreateServiceClient = networking::service::Client<PlatoonService>::create(
        net, PlatoonMessage::MAX_SIZE);
}

void Vehicle::addVehicleEndpoint(VehicleId vehicleId, const std::string & host)
{
    if (vehicleId == 0)
        throw std::invalid_argument{"Vehicle id must not be 0!"};

    auto endpoint = createVehicleEndpoint(vehicleId, host);

    auto self = shared_from_this();
    net.callLater(
        [self, endpoint]
        { self->vehicleEndpoints.push_back(endpoint); });
}

bool Vehicle::isIdle()
{
    std::lock_guard<std::mutex> lock(stateMutex);
    return state == State::IDLE;
}

bool Vehicle::isPlatoonRunning()
{
    std::lock_guard<std::mutex> lock(stateMutex);
    return state == State::RUNNING_PLATOON;
}

void Vehicle::createPlatoon()
{
    auto self = shared_from_this();
    net.callLater(
        [self]
        { self->doCreatePlatoon(); });
}

void Vehicle::leavePlatoon()
{
    auto self = shared_from_this();
    net.callLater(
        [self]
        { self->doLeavePlatoon(); });
}

void Vehicle::stop()
{
    platoonCreateServiceServer->stop();

    platoonCreateServiceClient->stop();
}

void Vehicle::doCreatePlatoon()
{
    auto self = shared_from_this();

    try
    {
        platoonCreateServiceServer->advertiseService(
            [self](const auto & clientEndpoint, const auto & requestMessage, auto & responseMessage)
            { self->makePlatoonCreateResponse(clientEndpoint, requestMessage, responseMessage); });
    }
    catch (...)
    {
        // In case failure (which will be the case when the port is being used already) we just keep calling this again.
        net.callLater(
            [self]
            { self->doCreatePlatoon(); });
        return;
    }

    // We cycle through every available endpoint and try to make a create-platoon-request.
    vehicleEndpointsIter = vehicleEndpoints.begin();
    makePlatoonCreateRequestToNextEndpoint();
}

void Vehicle::makePlatoonCreateRequestToNextEndpoint()
{
    if (state != State::CREATING_PLATOON)
        return;

    if (!platoonCreateServiceClient->isCalling() &&
        vehicleEndpointsIter != vehicleEndpoints.end()) // in case of an empty vector of vehicle endpoints
    {
        auto vehicleEndpoint = *vehicleEndpointsIter;
        makePlatoonCreateRequest(vehicleEndpoint);
        utils::cycle(vehicleEndpointsIter, vehicleEndpoints);
    }

    auto self = shared_from_this();
    net.callLater(
        [self]
        { self->makePlatoonCreateRequestToNextEndpoint(); });
}

void Vehicle::setState(Vehicle::State newState)
{
    std::lock_guard<std::mutex> lock(stateMutex);
    state = newState;
}

VehicleEndpoint Vehicle::createVehicleEndpoint(VehicleId vehicleId, const std::string & host)
{
    if (vehicleId == 0)
        throw std::invalid_argument{"Vehicle id must not be 0!"};

    return VehicleEndpoint{
        vehicleId,
        host,
#ifdef PLATOON_PROTOCOL_ENABLE_LOCAL_TESTING
        // To test this locally, we must ensure that each vehicle has a different port.
        TCP_PORT + vehicleId,
        UDP_PORT + vehicleId
#else
    TCP_PORT,
    UDP_PORT
#endif
    };
}
